"""IIIF Image API module."""
from iiif._version import __version__

from iiif.error import IIIFError
from iiif.info import IIIFInfo
from iiif.manipulator import IIIFManipulator
from iiif.request import IIIFRequest
from iiif.static import IIIFStatic
